import configparser
import contextlib
import copy
import dataclasses
import datetime
import enum
import functools
import pathlib
import platform
import threading
import time
from typing import ClassVar, Literal, NamedTuple, NoReturn, Optional, TypedDict

import IPython
import IPython.display
import ipywidgets as ipw
import np_config
import np_logging
import np_services
import np_session
import np_workflows
import PIL.Image
import pydantic
from pyparsing import Any
from np_services import (
    Service,
    Finalizable,
    ScriptCamstim,
    OpenEphys,
    Sync,
    VideoMVR,
    NewScaleCoordinateRecorder,
    MouseDirector,
)

from .ttn_stim_config import (
    TTNSession,
    camstim_defaults,
    per_session_main_stim_params,
    per_session_mapping_params,
    per_session_opto_params,
    default_ttn_params as DEFAULT_STIM_PARAMS,
)

logger = np_logging.getLogger(__name__)


class TTNMixin:
    """Provides TTN-specific methods and attributes, mainly related to camstim scripts."""
    
    ttn_session: TTNSession
    """Enum for session type, e.g. PRETEST, HAB_60, HAB_90, ECEPHYS."""
    @property
    def recorders(self) -> tuple[Service, ...]:
        """Services to be started before stimuli run, and stopped after. Session-dependent."""
        match self.ttn_session:
            case TTNSession.PRETEST | TTNSession.ECEPHYS:
                return (Sync, VideoMVR, OpenEphys)
            case TTNSession.HAB_60 | TTNSession.HAB_90 | TTNSession.HAB_120:
                return (Sync, VideoMVR)

    @property
    def stims(self) -> tuple[Service, ...]:
        return (ScriptCamstim,)

    def initialize_and_test_services(self) -> None:
        """Configure, initialize (ie. reset), then test all services."""
        
        MouseDirector.user = self.user.id
        MouseDirector.mouse = self.mouse.id

        OpenEphys.folder = self.session.folder

        NewScaleCoordinateRecorder.log_root = self.session.npexp_path

        self.configure_services()

        super().initialize_and_test_services()

    def run_stim_scripts(self) -> None:
        for stim in ('mapping', 'main', 'opto'):
            
            if not (params := self.params[stim]):
                logger.info("%s script skipped this session: %r", stim, self.ttn_session)
                continue
            
            ScriptCamstim.params = params
            ScriptCamstim.script = self.scripts[stim]
            
            logger.info("Starting %s script", stim)
            ScriptCamstim.start()
            
            with contextlib.suppress(Exception):
                while not ScriptCamstim.is_ready_to_start():
                    time.sleep(2.5)
                
            logger.info("%s script complete", stim)
            
            if isinstance(ScriptCamstim, Finalizable):
                ScriptCamstim.finalize()

    @property
    def params(self) -> dict[Literal["main", "mapping", "opto", "system"], dict[str, Any]]:
        params = copy.deepcopy(DEFAULT_STIM_PARAMS)
        if system := self.system_camstim_params:
            params["system"] = system
        params["main"] = per_session_main_stim_params(self.ttn_session)
        params["mapping"] = per_session_mapping_params(self.ttn_session)
        params["opto"] =  per_session_opto_params(self.ttn_session, self.mouse)
        return params

    @functools.cached_property
    def scripts(self) -> dict[Literal["main", "mapping", "opto"], str]:
        "Relative path to script on Stim computer from current location."
        return {label: np_config.local_to_unc(
            platform.node(),
            pathlib.Path(__file__).parent / 'camstim_scripts' / f"ttn_{label}_script.py"
        ).as_posix() for label in ("main", "mapping", "opto")}

    @functools.cached_property
    def system_camstim_params(self) -> dict[str, Any]:
        "System config on Stim computer, if accessible."
        return camstim_defaults()
    
class Hab(TTNMixin, np_workflows.Hab):
    def __init__(self, *args, **kwargs):
        self.services = (
            MouseDirector,
            Sync,
            VideoMVR,
            self.imager,
            ScriptCamstim,
        )
        super().__init__(*args, **kwargs)


class Ecephys(TTNMixin, np_workflows.Ecephys):
    def __init__(self, *args, **kwargs):
        self.services = (
            MouseDirector,
            Sync,
            VideoMVR,
            self.imager,
            ScriptCamstim,
            OpenEphys,
            NewScaleCoordinateRecorder,
        )
        super().__init__(*args, **kwargs)


# --------------------------------------------------------------------------------------


def new_experiment(
    mouse: int | str | np_session.Mouse,
    user: str | np_session.User,
    session: TTNSession,
) -> Ecephys | Hab:
    """Create a new experiment for the given mouse and user."""
    match session:
        case TTNSession.PRETEST | TTNSession.ECEPHYS:
            experiment = Ecephys(mouse, user)
        case TTNSession.HAB_60 | TTNSession.HAB_90 | TTNSession.HAB_120:
            experiment = Hab(mouse, user)
        case _:
            raise ValueError(f"Invalid session type: {session}")
    experiment.ttn_session = session
    logger.info("Created new experiment session: %s", experiment)
    return experiment


# --------------------------------------------------------------------------------------




import datetime
import time
from typing import Type

import np_logging
import np_session

from np_workflows.experiments import baseclasses, classes
from np_workflows.experiments.baseclasses import Experiment
# we need all experiment classes imported into this namespace
from np_workflows.experiments.classes import *
from np_workflows.workflows.shared import npxc

logger = np_logging.getLogger(__name__)


# mouse & operator ---------------------------------------------------------------------


def select_mouse_and_operator_enter(state_globals) -> None:
    npxc.set(state_globals, operators=[""] + npxc.get_operators())
    state_globals["external"]["mouse"] = "366122"  #! defaults for testing
    # state_globals["external"]["operator"] = 'ben.hardcastle' #! defaults for testing

    # import pdb; pdb.set_trace()
    # state_globals['resources']['workflow']['states'].keys()


def select_mouse_and_operator_input(state_globals) -> None:

    mouse_input = state_globals["external"]["mouse"]
    try:
        npxc.mouse = np_session.lims.MouseInfo(mouse_input)
    except KeyError as exc:
        logger.warning(
            "LIMS | Mouse %r could not be fetched from database: %r", mouse_input, exc
        )
        npxc.mouse = mouse_input
        npxc.is_mouse_in_lims = False

    try:
        project = npxc.mouse.project_name
    except (AttributeError, KeyError):
        project = "NA"
        if npxc.is_mouse_in_lims is None:
            logger.warning(
                "LIMS | Mouse %r could not be fetched from database: %r", mouse_input
            )
        npxc.mouse = mouse_input
        npxc.is_mouse_in_lims = False
    else:
        npxc.is_mouse_in_lims = True
    finally:
        npxc.set(
            state_globals, project=str(project), is_mouse_in_lims=npxc.is_mouse_in_lims
        )

    operator_input = state_globals["external"]["operator"]
    try:
        npxc.operator = np_session.lims.UserInfo(operator_input)
    except KeyError as exc:
        logger.warning(
            "LIMS | Operator %r could not be fetched from database: %r",
            operator_input,
            exc,
        )
        npxc.operator = operator_input

    status_message = f'Mouse {"is" if npxc.is_mouse_in_lims else "is not"} in lims'
    transition_result = (
        False if not npxc.is_mouse_in_lims else True
    )  # None isn't allowed in s_g['external']
    npxc.set(
        state_globals,
        status_message=status_message,
        transition_result=transition_result,
    )


# experiment ---------------------------------------------------------------------------


def select_experiment_enter(state_globals) -> None:
    experiments = npxc.get_experiments(with_lims=bool(npxc.is_mouse_in_lims))
    if not experiments:
        alert = True
        msg_text = "No experiments found"
        npxc.set(state_globals, alert=alert, msg_text=msg_text)
    npxc.set(state_globals, experiments=[""] + experiments)


def select_experiment_input(state_globals) -> None:
    "Create Experiment instance and configure services"

    experiment_class: Type[Experiment] = eval(npxc.get(state_globals, "experiment"))

    if issubclass(experiment_class, baseclasses.WithLims):
        npxc.experiment = experiment_class(
            *npxc.get(state_globals, "mouse", "operator"),
            lims_session_id=int(npxc.session) if npxc.session else None,
        )
    else:
        npxc.experiment = experiment_class(npxc.mouse)

    if not isinstance(npxc.experiment, Experiment):
        raise ValueError(
            f"Experiment {npxc.experiment!r} was not converted to a valid Experiment class"
        )

    npxc.experiment.state_globals = state_globals
    npxc.experiment.initialize_services()



import configparser
import copy
import enum
import functools
from typing import ClassVar, Literal, NamedTuple, NoReturn, Optional, TypedDict

import IPython.display
import ipywidgets as ipw
import np_config
import np_logging
import np_session
import np_workflows
from pyparsing import Any

from .ttn_stim_config import TTNSession

# for widget, before creating a experiment --------------------------------------------- #


class TTNSelectedSession:
    def __init__(self, session: str | TTNSession, mouse: str | int | np_session.Mouse):
        if isinstance(session, str):
            session = TTNSession(session)
        self.session = session
        self.mouse = str(mouse)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.session}, {self.mouse})"


def stim_session_select_widget(
    mouse: str | int | np_session.Mouse,
) -> TTNSelectedSession:
    """Select a stimulus session (hab, pretest, ecephys) to run.

    An object with mutable attributes is returned, so the selected session can be
    updated along with the GUI selection. (Preference would be to return an enum
    directly, and change it's value, but that doesn't seem possible.)

    """

    selection = TTNSelectedSession(TTNSession.PRETEST, mouse)

    session_dropdown = ipw.Select(
        options=tuple(_.value for _ in TTNSession),
        description="Session",
    )
    console = ipw.Output()
    with console:
        print(f"Selected: {selection.session}")

    def update(change):
        if change["name"] != "value":
            return
        if (options := getattr(change["owner"], "options", None)) and change[
            "new"
        ] not in options:
            return
        if change["new"] == change["old"]:
            return
        selection.__init__(str(session_dropdown.value), mouse)
        with console:
            print(f"Selected: {selection.session}")

    session_dropdown.observe(update)

    IPython.display.display(ipw.VBox([session_dropdown, console]))

    return selection



import contextlib
import datetime
import logging
import pathlib
import threading
import time
from typing import NoReturn, Optional

import IPython
import IPython.display
import ipywidgets as ipw
import np_config
import np_logging
import np_session
import np_services
import PIL.Image

import np_workflows.npxc as npxc


def elapsed_time_widget() -> IPython.display.DisplayHandle | None:
    """Displays a clock showing the elapsed time since the cell was first run."""

    clock_widget = ipw.Label("")
    reminder_widget = ipw.Label("Remember to restart the kernel for every experiment!")
    global start_time
    if "start_time" not in globals():
        start_time = time.time()

    if isinstance(start_time, datetime.datetime):
        start_time = start_time.timestamp()

    def update_timer() -> NoReturn:
        while True:
            elapsed_sec = time.time() - start_time
            hours, remainder = divmod(elapsed_sec, 3600)
            minutes, seconds = divmod(remainder, 60)
            clock_widget.value = "Elapsed time: {:02}h {:02}m {:02}s".format(
                int(hours), int(minutes), int(seconds)
            )
            if hours > 4:  # ipywidgets >= 8.0
                clock_widget.style = dict(text_color="red",)
            time.sleep(0.2)

    thread = threading.Thread(target=update_timer, args=())
    thread.start()
    return IPython.display.display(ipw.VBox([clock_widget, reminder_widget]))


def user_and_mouse_widget() -> tuple[np_session.User, np_session.Mouse]:
    user_description = "User:"
    mouse_description = "Mouse:"
    user_widget = ipw.Select(options=npxc.lims_user_ids, description=user_description)
    mouse_widget = ipw.Text(value=str(npxc.default_mouse_id), description=mouse_description)
    user = np_session.User(str(user_widget.value))
    mouse = np_session.Mouse(str(mouse_widget.value))
    console = ipw.Output()

    def update_user(new_user: str):
        if str(user) == (new := str(new_user).strip()):
            return
        user.__init__(new)
        with console:
            print(f"User updated: {user}")

    def update_mouse(new_mouse: str):
        if str(mouse) == (new := str(new_mouse).strip()):
            return
        if len(new) < 6:
            return
        mouse.__init__(new)
        with console:
            print(f"Mouse updated: {mouse}")
        
    def new_value(change) -> None:
        if change['name'] != 'value':
            return
        if (options := getattr(change['owner'], 'options', None)) and change['new'] not in options:
            return
        if change['new'] == change['old']:
            return
        if (desc := getattr(change['owner'], 'description')) == user_description:
                update_user(change['new'])
        elif desc == mouse_description:
                update_mouse(change['new'])
            
    user_widget.observe(new_value)
    mouse_widget.observe(new_value)
    
    IPython.display.display(ipw.VBox([user_widget, mouse_widget, console]))
    return user, mouse


def mtrain_widget(
    labtracks_mouse_id: str | int | np_session.Mouse,
) -> IPython.display.DisplayHandle | None:
    """Displays a widget to view and edit MTrain regimen/stage for a mouse.
    """
    if not isinstance(labtracks_mouse_id, np_session.Mouse):
        mtrain = np_session.MTrain(labtracks_mouse_id)
    else:
        mtrain = labtracks_mouse_id.mtrain

    all_regimens = mtrain.get_all("regimens")
    regimen_names = sorted(_["name"] for _ in all_regimens)

    widget = ipw.GridspecLayout(n_rows=4, n_columns=2)

    # labels
    widget[0, 0] = ipw.Label(f"Mouse: {mtrain.mouse_id}")
    widget[1, 0] = regimen_label = ipw.Label("Regimen:")
    widget[2, 0] = stage_label = ipw.Label("Stage:")

    # dropdowns
    widget[1, 1] = regimen_dropdown = ipw.Dropdown(options=regimen_names)
    widget[2, 1] = stage_dropdown = ipw.Dropdown(
        options=sorted([_["name"] for _ in mtrain.stages])
    )
    stage_dropdown.stages: list[dict] = mtrain.stages

    widget[3, 1] = update_button = ipw.Button(description="Update", disabled=True)

    console = ipw.Output()

    display = ipw.VBox([widget, console])

    def on_regimen_change(change: dict):
        update_button.disabled = True
        new_regimen_dict = [
            regimen
            for regimen in all_regimens
            if regimen["name"] == regimen_dropdown.value
        ][0]
        stage_dropdown.options = sorted([_["name"] for _ in new_regimen_dict["stages"]])
        stage_dropdown.value = None
        stage_dropdown.stages = new_regimen_dict["stages"]

    regimen_dropdown.observe(on_regimen_change, names="value")

    def reset_update_button():
        update_button.description = "Update"
        update_button.disabled = True
        update_button.button_style = ""

    def on_stage_change(change: dict):
        reset_update_button()
        if change["new"] is None:
            return
        if change["new"] != stage_label.value or str(
            regimen_dropdown.value
        ) != str(regimen_label.value):
            # enable button if stage name changed, or regimen name changed (some
            # regimens have the same stage names as other regimens)
            update_button.disabled = False
            update_button.button_style = "warning"

    stage_dropdown.observe(on_stage_change, names="value")

    def update_label_values() -> None:
        regimen_label.value = f'Regimen: {mtrain.regimen["name"]}'
        stage_label.value = f'Stage: {mtrain.stage["name"]}'

    def update_dropdown_values() -> None:
        regimen_dropdown.value = mtrain.regimen["name"]
        stage_dropdown.value = mtrain.stage["name"]

    def update_regimen_and_stage_in_mtrain(b):
        update_button.description = "Updating..."
        update_button.disabled = True

        old_regimen_name = regimen_label.value
        old_stage_name = stage_label.value

        new_regimen_dict = [
            _ for _ in all_regimens if _["name"] == regimen_dropdown.value
        ][0]
        new_stage_dict = [
            _ for _ in stage_dropdown.stages if _["name"] == stage_dropdown.value
        ][0]

        mtrain.set_regimen_and_stage(regimen=new_regimen_dict, stage=new_stage_dict)
        update_all()

        regimen_name_changed: bool = new_regimen_dict["name"] not in old_regimen_name
        stage_name_changed: bool = new_stage_dict["name"] not in old_stage_name
        with console:
            if regimen_name_changed:
                print(f'{old_regimen_name} changed to {mtrain.regimen["name"]}\n')
            if stage_name_changed or regimen_name_changed:
                print(f'{old_stage_name} changed to {mtrain.stage["name"]}\n')

    update_button.on_click(update_regimen_and_stage_in_mtrain)

    def update_all():
        update_label_values()
        update_dropdown_values()
        reset_update_button()
        update_label_values()
        update_dropdown_values()

    update_all()

    return IPython.display.display(display)

def check_widget(check: str, *checks: str) -> ipw.Widget:
    layout = ipw.Layout(min_width="600px")
    widget = ipw.VBox([
        ipw.Label(check, layout=layout), 
        *(ipw.Checkbox(description=_, layout=layout) for _ in checks),
        # ipw.Button(description="Continue", disabled=True)
        ])
    return widget

def await_all_checkboxes(widget: ipw.Box) -> None:
    while any(_.value == False for _ in widget.children if isinstance(_, ipw.Checkbox)):
        time.sleep(0.1)
    
    
def check_openephys_widget() -> None:
    check = "OpenEphys checks:"
    checks = (
        "Record Node paths are set to two different drives (A: & B: or E: & G:)",
        "Each Record Node recording only ABC or DEF probes",
        "Tip-reference on all probes",
        "Barcodes visible",
    )
    IPython.display.display(widget := check_widget(check, *checks))

def check_hardware_widget() ->  None:
    check = "Stage checks:"
    checks = (
        "Cartridge raised (fully retract probes before raising!)",
        "Water lines flushed",
    )
    IPython.display.display(widget := check_widget(check, *checks))

def check_mouse_widget() -> None:
    check = "Mouse checks before lowering cartridge:"
    checks = (
        "Stabilization screw",
        "Quickcast, agarose, silicon oil",
        "Tail cone down",
        "Continuity/Resistance check",
        "Eye-tracking mirror in place",
    )
    IPython.display.display(widget := check_widget(check, *checks))

def pre_stim_check_widget() -> None:
    check = "Before running stim:"
    checks = (
        "Windows minimized on Stim computer (Win+D)",
        "Eye-tracking mirror clean", 
        "Monitor closed",
        "Photodoc light off",
        "Curtain down",     
    )
    IPython.display.display(widget := check_widget(check, *checks))
    
def finishing_checks_widget() -> None:
    check = "Finishing checks:"
    checks = (
        "Add quickcast etc.",
        "Remove and water mouse", 
        "Dip probes",     
    )
    IPython.display.display(widget := check_widget(check, *checks))
    
    
def wheel_height_widget(platform_json: pathlib.Path | np_services.PlatformJsonWriter) -> IPython.display.DisplayHandle | None:
    "Supply a path or a platform json instance. Saves a JSON file with the wheel height recorded."

    if isinstance(platform_json, pathlib.Path):
        platform_json = np_services.PlatformJsonWriter(path=platform_json)
    platform_json.load_from_existing()
    
    if platform_json.mouseID:
        mouse = np_session.Mouse(platform_json.mouseID)
        
    layout = ipw.Layout(max_width='130px')
    
    try:
        prev_height = mouse.state.get('wheel_height', 0)
    except:
        prev_height = 0
    height_counter = ipw.BoundedFloatText(value=prev_height, min=0, max=10, step=0.1, description="Wheel height", layout=layout)
    save_button = ipw.Button(description='Save', button_style='warning', layout=layout)

    def on_click(b):
        platform_json.load_from_existing()
        platform_json.wheel_height = height_counter.value
        with contextlib.suppress(Exception):
            mouse.state['wheel_height'] = height_counter.value
        platform_json.write(update_existing=False)
        save_button.button_style = 'success'
        save_button.description = 'Saved'
    save_button.on_click(on_click)
    return IPython.display.display(ipw.VBox([height_counter,save_button]))
    
    
def di_widget(platform_json: pathlib.Path | np_services.PlatformJsonWriter) -> IPython.display.DisplayHandle | None:
    "Supply a path or a platform json instance. Saves a JSON file with the dye used in the session and a timestamp."

    di_info: dict[str, int | str] = dict(
        EndTime=0, StartTime=str(npxc.now()), dii_description="", times_dipped=0,
    )
    if isinstance(platform_json, pathlib.Path):
        platform_json = np_services.PlatformJsonWriter(path=platform_json)
    layout = ipw.Layout(max_width='130px')
    dipped_counter = ipw.IntText(value=0, min=0, max=99, description="Dipped count", layout=layout)
    usage_counter = ipw.IntText(value=0, min=0, max=99, description="Previous uses", layout=layout)
    dye_dropdown = ipw.Dropdown(options=['CM-DiI 100%', 'DiO'], layout=layout)
    save_button = ipw.Button(description='Save', button_style='warning', layout=layout)
    
    def update_di_info():
        di_info['EndTime'] = str(npxc.now())
        di_info['times_dipped'] = str(dipped_counter.value)
        di_info['dii_description'] = str(dye_dropdown.value)
        di_info['previous_uses'] = str(usage_counter.value)
        
    def on_click(b):
        platform_json.load_from_existing()
        update_di_info()
        platform_json.DiINotes = di_info
        platform_json.write(update_existing=False)
        save_button.button_style = 'success'
        save_button.description = 'Saved'
    save_button.on_click(on_click)
    return IPython.display.display(ipw.VBox([
        dipped_counter, dye_dropdown, 
        usage_counter, save_button]))

    
def dye_widget(session_folder: pathlib.Path) -> IPython.display.DisplayHandle | None:
    "Supply a path - saves a JSON file with the dye used in the session and a timestamp."

    di_info: dict[str, int | str] = dict(
        EndTime=0, StartTime=0, dii_description="DiI", times_dipped=0,
    )
        
    class DyeRecorder(np_services.JsonRecorder):
        log_name = f'{session_folder.name}_dye.json'
        log_root = session_folder

    dye_dropdown = ipw.Dropdown(options=['DiI', 'DiO'])
    save_button = ipw.Button(description='Save', button_style='warning')
    def on_click(b):
        DyeRecorder.write(dict(dye=dye_dropdown.value, datetime=datetime.datetime.now(), time=time.time()))
        save_button.button_style = 'success'
        save_button.description = 'Saved'
    save_button.on_click(on_click)
    return IPython.display.display(ipw.VBox([dye_dropdown, save_button]))

def isi_widget(
    labtracks_mouse_id: str | int | np_session.LIMS2MouseInfo, colormap: bool = False,
) -> IPython.display.DisplayHandle | None:
    """Displays ISI target map from lims (contours only), or colormap overlay if
    `show_colormap = True`."""
    if not isinstance(labtracks_mouse_id, np_session.LIMS2MouseInfo):
        lims_info = np_session.LIMS2MouseInfo(labtracks_mouse_id)
    else:
        lims_info = labtracks_mouse_id

    if colormap:
        key = "isi_image_overlay_path"
    else:
        key = "target_map_image_path"
    
    try:
        lims_path = lims_info.isi_info[key]
    except ValueError:
        print("Mouse is not in lims.")
        return
    except (AttributeError, TypeError):
        print("No ISI map found for this mouse.")
        return
    except KeyError:
        print(f"ISI info found for this mouse, but {key=!r} is missing.")
        return IPython.display.display(IPython.display.JSON(lims_info.isi_info))
    else:
        path: pathlib.Path = np_config.normalize_path(lims_path)
        print(f"ISI map found for {lims_info.np_id}:\n{path}")
        img = PIL.Image.open(path)
        return IPython.display.display(img)
    
    # alternative to use ipw with toggle button -------------------------------------------- #
    # import io
    # membuf = io.BytesIO()
    # img.save(membuf, format="png") 
    # return IPython.display.display(ipw.VBox([ipw.Image(value=membuf.getvalue())]))


def photodoc_widget(img_name: str) -> IPython.display.DisplayHandle | None:
    "Captures and displays snapshot from image camera, appending `img_name` to the filename."
    image = ipw.Image(value=b'', format='png', width='80%', layout=ipw.Layout(visibility='hidden'))
    widget = ipw.VBox([
        image,
        button := ipw.Button(description="Capture", button_style='warning'),
        console := ipw.Output(),
    ])
    def capture() -> pathlib.Path:
        image.value = b''
        image.layout.visibility = 'hidden'
        button.button_style = ''
        button.description = 'Capturing image...'
        button.disabled = True
        return npxc.photodoc(img_name)
    def capture_and_display(b):
        img_path = capture()
        # image.value = PIL.Image.open(img_path).tobytes()
        image.value = img_path.read_bytes()
        image.layout.visibility = 'visible'
        button.button_style = 'warning'
        button.description = 'Capture'
        button.disabled = False
        with console:
            print(img_path)
    button.on_click(capture_and_display)
    return IPython.display.display(widget)

def probe_targeting_widget(session_folder) -> IPython.display.DisplayHandle | None:
    from np_probe_targets.implant_drawing import CurrentWeek, DRWeeklyTargets
    CurrentWeek.display()
    IPython.display.display(DRWeeklyTargets())
    
def quiet_mode_widget() -> IPython.display.DisplayHandle | None:
    """Displays a toggle button that switches logging level INFO <-> DEBUG and
    hides/shows tracebacks.
    """
    debug_mode_toggle = ipw.ToggleButton(
            value=True,
            description='Quiet mode is on',
            disabled=False,
            button_style='info', # 'success', 'info', 'warning', 'danger' or ''
            icon='check',
            tooltip='Quiet mode: tracebacks hidden, logging level set to INFO.',
        )
    def set_debug_mode(value: bool) -> None:
        if value:
            npxc.show_tracebacks()
            for handler in np_logging.getLogger().handlers:
                if isinstance(handler, logging.StreamHandler):
                    handler.setLevel('DEBUG')
        else:
            npxc.hide_tracebacks()
            for handler in np_logging.getLogger().handlers:
                if isinstance(handler, logging.StreamHandler):
                    handler.setLevel('INFO')
            
    np_logging.getLogger('Comm').setLevel('INFO')
    
    def on_click(b) -> None:
        if not debug_mode_toggle.value:
            set_debug_mode(True)
            debug_mode_toggle.description = 'Quiet mode is off'
            debug_mode_toggle.button_style = ''
            debug_mode_toggle.icon = 'times'
        else:
            set_debug_mode(False)
            debug_mode_toggle.description = 'Quiet mode is on'
            debug_mode_toggle.button_style = 'info'
            debug_mode_toggle.icon = 'check'
    
    debug_mode_toggle.observe(on_click)
     
    return IPython.display.display(debug_mode_toggle)



import configparser
import contextlib
import copy
import enum
import functools
from typing import ClassVar, Literal, NamedTuple, NoReturn, Optional, TypedDict

import IPython.display
import ipywidgets as ipw
import np_config
import np_logging
import np_session
import np_workflows
from typing import Any

logger = np_logging.getLogger(__name__)


class TTNSession(enum.Enum):
    """Enum for the different TTN sessions available, each with different param sets."""

    PRETEST = "pretest"
    HAB_60 = "hab 60"
    HAB_90 = "hab 90"
    HAB_120 = "hab 120"
    ECEPHYS = "ecephys"


# setup parameters ---------------------------------------------------------------------
default_ttn_params = {}

def camstim_defaults() -> dict:
    """Try to load defaults from camstim config file on the Stim computer.
    
    May encounter permission error if not running as svc_neuropix.
    """
    with contextlib.suppress(OSError):
        parser = configparser.RawConfigParser()
        parser.read(
            (np_config.Rig().paths["Camstim"].parent / "config" / "stim.cfg").as_posix()
        )

        camstim_default_config = {}
        for section in parser.sections():
            camstim_default_config[section] = {}
            for k, v in parser[section].items():
                try:
                    value = eval(
                        v
                    )  # this removes comments in config and converts values to expected datatype
                except:
                    continue
                else:
                    camstim_default_config[section][k] = value
        return camstim_default_config
    logger.warning("Could not load camstim defaults from config file on Stim computer.")
    return {}

## no longer added to default_ttn_params:
# default_ttn_params.update(**camstim_defaults())


# main stimulus defaults ---------------------------------------------------------------
default_ttn_params["main"] = {}

default_ttn_params["main"]["sweepstim"] = {
    "trigger_delay_sec": 5.0,       #! does it matter?
    'sync_sqr_loc': (870, 525),     # for stim with warp=Warp.disabled 
}
# default_ttn_params["main"]["movie_path"] = "C:/ProgramData/StimulusFiles/dev/"
default_ttn_params["main"][
    "monitor"
] = 'Gamma1.Luminance50'

# other parameters that vary depending on session type (pretest, hab, ecephys):
def per_session_main_stim_params(session: TTNSession) -> dict[str, Any]:
    def build_session_stim_params(
        key: str, old: int, reversed: int, annotated: int
    ) -> dict[str, dict[str, int]]:
        return {key: dict(old=old, reversed=reversed, annotated=annotated)}

    def session_stim_repeats(session: TTNSession) -> dict[str, dict[str, int]]:
        repeats = functools.partial(build_session_stim_params, "stim_repeats")
        match session:
            case TTNSession.PRETEST:
                return repeats(1, 1, 1)
            case TTNSession.HAB_60:
                return repeats(15, 5, 1)
            case TTNSession.HAB_90:
                return repeats(20, 7, 2)
            case TTNSession.HAB_120 | TTNSession.ECEPHYS:
                return repeats(25, 8, 2)
            case _:
                raise ValueError(f"Stim repeats not implemented for {session}")

    def session_stim_lengths(session: TTNSession) -> dict[str, dict[str, int]]:
        lengths_sec = functools.partial(build_session_stim_params, "stim_lengths_sec")
        match session:
            case TTNSession.PRETEST:
                return lengths_sec(1, 1, 1)
            case _:
                return lengths_sec(40, 40, 60)

    def main_blank_screen(session: TTNSession) -> dict[str, float | int]:
        match session:
            case TTNSession.PRETEST:
                return {"pre_blank_screen_sec": .5, "post_blank_screen_sec": .5}
            case _:
                return {"pre_blank_screen_sec": 2, "post_blank_screen_sec": 2}

    params = copy.deepcopy(default_ttn_params["main"])
    params.update(session_stim_repeats(session))
    params.update(session_stim_lengths(session))
    params.update(main_blank_screen(session))
    return params


# optotagging defaults -----------------------------------------------------------------

default_ttn_params["opto"] = {}

# all parameters depend on session type (pretest, hab, ecephys):

def per_session_opto_params(
    session: TTNSession, mouse: str | int | np_session.Mouse
) -> dict[str, dict[str, str | list[float] | Literal["pretest", "experiment"]]]:
    "All params for opto depending on session (e.g. will be empty for habs)."

    def opto_mouse_id(mouse_id: str | int | np_session.Mouse) -> dict[str, str]:
        return {"mouseID": str(mouse_id)}

    def opto_levels(session: TTNSession) -> dict[str, list[float]]:
        default_opto_levels: list[float] = camstim_defaults()["Optogenetics"][
            "level_list"
        ]
        match session:
            case TTNSession.PRETEST | TTNSession.ECEPHYS:
                return {"level_list": sorted(default_opto_levels)[-2:]}
            case _:
                raise ValueError(f"Opto levels not implemented for {session}")

    def opto_operation_mode(
        session: TTNSession,
    ) -> dict[str, Literal["pretest", "experiment"]]:
        match session:
            case TTNSession.PRETEST:
                return {"operation_mode": "pretest"}
            case TTNSession.ECEPHYS:
                return {"operation_mode": "experiment"}
            case _:
                raise ValueError(f"Opto levels not implemented for {session}")

    match session:
        case TTNSession.PRETEST | TTNSession.ECEPHYS:
            params = copy.deepcopy(default_ttn_params["opto"])
            params.update(opto_mouse_id(mouse))
            params.update(opto_levels(session))
            params.update(opto_operation_mode(session))
            return params
        case TTNSession.HAB_60 | TTNSession.HAB_90 | TTNSession.HAB_120:
            return {}
        case _:
            raise ValueError(f"Opto params not implemented for {session}")


# mapping defaults ---------------------------------------------------------------------

default_ttn_params["mapping"] = {}

default_ttn_params["mapping"]["monitor"] = "Gamma1.Luminance50"
default_ttn_params["mapping"]["gabor_path"] = "gabor_20_deg_250ms.stim" 
default_ttn_params["mapping"]["flash_path"] = "flash_250ms.stim" # relpath in StimulusFiles dir (for _v2 objects)
default_ttn_params["mapping"]["default_gabor_duration_seconds"] = 1200
default_ttn_params["mapping"]["default_flash_duration_seconds"] = 300 # may be overriden by 'max_total_duration_minutes'

default_ttn_params["mapping"]["sweepstim"] = {} 
# trigger_delay_sec not specified
# all stim parameters depend on session type (pretest, hab, ecephys):
def per_session_mapping_params(session: TTNSession) -> dict[str, dict[str, int]]:
    "`'mapping'` key in params dict should be updated with the returned dict (which will be empty for habs)."

    def mapping_duration(session: TTNSession) -> dict[str, float]:
        # 0 = full length = gabor_duration + flash_duration = maximum possible
        match session:
            case TTNSession.PRETEST:
                return {"max_total_duration_minutes": 0.1}
            case TTNSession.ECEPHYS | TTNSession.HAB_120:
                return {"max_total_duration_minutes": 10}
            case _:
                raise ValueError(f"Mapping params not implemented for {session}")

    def mapping_blank_screen(session: TTNSession) -> dict[str, float | int]:
        match session:
            case TTNSession.PRETEST:
                return {"pre_blank_screen_sec": .5, "post_blank_screen_sec": .5}
            case TTNSession.ECEPHYS | TTNSession.HAB_120:
                return {"pre_blank_screen_sec": 2, "post_blank_screen_sec": 2}
            case _:
                raise ValueError(f"Mapping params not implemented for {session}")

    match session:
        case TTNSession.PRETEST | TTNSession.ECEPHYS | TTNSession.HAB_120:
            params = copy.deepcopy(default_ttn_params["mapping"])
            params.update(mapping_duration(session))
            params.update(mapping_blank_screen(session))
            return params
        case TTNSession.HAB_60 | TTNSession.HAB_90:
            return {}
        case _:
            raise ValueError(f"Mapping params not implemented for {session}")



"""
Oct'22 task-trained ephys stimuli
"""

import argparse
import json
import logging
import os
import time

import numpy as np
from psychopy import visual
from camstim import Foraging
from camstim import Stimulus_v2
from camstim import SweepStim_v2
from camstim import Warp, Window


# get params ------------------------------------------------------------------
# stored in json file -
# path to json supplied by camstim via command line arg when this script is called

parser = argparse.ArgumentParser()
parser.add_argument(
    "params_path",
    nargs="?",
    type=str,
    default="",
)
args, _ = parser.parse_known_args()

with open(args.params_path, "r") as f:
    json_params = json.load(f)

# Create display window
# ----------------------------------------------------------------------------
window = Window(
    fullscr=True,
    monitor=json_params["monitor"],
    screen=0,
    warp=Warp.Spherical,
)

# patch the Stimulus_v2 class to allow for serializing without large arrays
# ----------------------------------------------------------------------------
class Stimulus_v2_MinusFrameArrays(Stimulus_v2):

    def __init__(self, *args, **kwargs):
        super(Stimulus_v2_MinusFrameArrays, self).__init__(*args, **kwargs)

    def package(self):
        """
        Package for serializing - minus large arrays of frame timing/order.
        """
        if not self.save_sweep_table:
            self.sweep_table = None
            self.sweep_params = self.sweep_params.keys()
        self_dict = self.__dict__
        del self_dict['sweep_frames']
        del self_dict['sweep_order']
        self_dict['stim'] = str(self_dict['stim'])
        return wecanpicklethat(self_dict)
    
# ----------------------------------------------------------------------------
# setup mapping stim
"""from mapping_script_v2.py"""

mapping_stimuli = []

# load common stimuli
gabor_path = json_params["gabor_path"]
flash_path = json_params["flash_path"]
gabor = Stimulus_v2_MinusFrameArrays.from_file(gabor_path, window)
flash = Stimulus_v2_MinusFrameArrays.from_file(flash_path, window)

gabor_duration_sec = json_params["default_gabor_duration_seconds"]
flash_duration_sec = json_params["default_flash_duration_seconds"]

original_duration_sec = gabor_duration_sec + flash_duration_sec

# if max total duration is set, and less than original movie length, cut down display sequence:
max_mapping_duation_minutes = json_params[
    "max_total_duration_minutes"
]  # can be zero, in which case we use the full movie length
max_mapping_duration_sec = max_mapping_duation_minutes * 60
if 0 < max_mapping_duration_sec < original_duration_sec:
    logging.info("Mapping duration capped at %s minutes", max_mapping_duation_minutes)

    logging.info("original gabor duration: %s sec", gabor_duration_sec)
    logging.info("original flash duration: %s sec", flash_duration_sec)
    logging.info("max mapping duration: %s sec", max_mapping_duration_sec)

    gabor_duration_sec = (
        max_mapping_duration_sec * gabor_duration_sec
    ) / original_duration_sec
    flash_duration_sec = (
        max_mapping_duration_sec * flash_duration_sec
    ) / original_duration_sec

    logging.info("modified gabor duration: %s sec", gabor_duration_sec)
    logging.info("modified flash duration: %s sec", flash_duration_sec)

# setup timing
mapping_sequence_start_sec = 0  # if stims are daisy-chained within one script, this should be the end of the prev stim
gabor.set_display_sequence([(mapping_sequence_start_sec, gabor_duration_sec)])
flash.set_display_sequence(
    [(gabor_duration_sec, gabor_duration_sec + flash_duration_sec)]
)

mapping_stimuli = [gabor, flash]

mapping_sequence_end_sec = (
    gabor_duration_sec + flash_duration_sec
)  # if daisy-chained, the next stim in this script should start at this time

# create SweepStim_v2 instance for main stimulus
ss = SweepStim_v2(
    window,
    stimuli=mapping_stimuli,
    pre_blank_sec=json_params["pre_blank_screen_sec"],
    post_blank_sec=json_params["post_blank_screen_sec"],
    params=json_params["sweepstim"],
)

# add in foraging so we can track wheel, potentially give rewards, etc
f = Foraging(
    window=window,
    auto_update=False,
    params=json_params["sweepstim"],
    nidaq_tasks={
        "digital_input": ss.di,
        "digital_output": ss.do,
    },
)  # share di and do with SS

ss.add_item(f, "foraging")

ss.run()



import contextlib
import datetime
import inspect
import pathlib
import shutil
import sys
import time
from typing import Any, Generator, Sequence, Type

import fabric
import np_config
import np_logging
import np_session
# from np_workflows.models import baseclasses, classes
# from np_workflows.models.baseclasses import Experiment
# from np_workflows.models import utils
from np_services import Initializable, Testable, TestError, Finalizable, Service
from np_services import Sync, VideoMVR, ImageMVR, Cam3d, ScriptCamstim, OpenEphys  
import np_services 
from np_config import Rig

logger = np_logging.getLogger(__name__)

# Assign default values to global variables so they can be imported elsewhere
# experiment: Experiment | str = ''
operator: np_session.User | str = ""
mouse: np_session.Mouse | str = ""
is_mouse_in_lims: bool | None = None
session: np_session.Session | str | int = ""

# experiments: tuple[Type[Experiment], ...] = (
#     classes.Pretest,
#     classes.NpUltra,
#     ) # TODO plug-in experiments

RIG = np_config.Rig()
CONFIG = RIG.config

lims_user_ids: tuple[str, ...] = tuple(
    sorted(
        CONFIG.get('lims_user_ids',
            [
            "hannah.belski",
            "hannah.cabasco",
            "ryan.gillis",
            "henry.loeffler",
            "corbettb",
            "ben.hardcastle",
            "samg",
            "ethan.mcbride",
            "jackie.kuyat",
            "andrew.shelton",
            ]
        )
    )
)

default_mouse_id: int = int(CONFIG.get('default_mouse_id', 366122))


def get_operators() -> list[str]:
    return list(lims_user_ids)


def get_experiments(with_lims: bool = True) -> list[str]:
    return [
        cls.__name__
        for cls in experiments
        if with_lims or not issubclass(cls, baseclasses.WithLims)
    ]


def initialize_services() -> None:
    experiment.apply_config_to_services()
    for service in experiment.services:

        if isinstance(service, Initializable):
            while True:
                try:
                    service.initialize()

                except Exception as exc:
                    logger.error("%s | %r", service.__name__, exc)
                    import pdb

                    pdb.set_trace()
                    continue

                else:
                    break

        if isinstance(service, Testable):
            while True:
                try:
                    service.test()

                except TestError as exc:
                    try:
                        logger.error("%s | %r", service.__name__, service.exc)
                    except AttributeError:
                        logger.error("%s | %r", service.__name__, exc)
                    import pdb

                    pdb.set_trace()
                    continue

                except Exception as exc:
                    logger.error("%s | %r", service.__name__, exc)
                    import pdb

                    pdb.set_trace()
                    continue

                else:
                    break


def print_countdown_timer(seconds: int | float | datetime.timedelta = 0, **kwargs):
    """Block execution for a given number of seconds (or any timedelta kwargs), printing a countdown timer to the console."""
    if isinstance(seconds, datetime.timedelta):
        wait = seconds
    else:
        wait = datetime.timedelta(seconds=seconds, **kwargs)
    time_0: float = time.time()
    time_remaining = lambda: datetime.timedelta(
        seconds=wait.total_seconds() - (time.time() - time_0)
    )
    while time_remaining().total_seconds() > 0:
        print(f"Waiting {wait} \t{time_remaining()}", end="\r", flush=True)
        time.sleep(0.1)

def photodoc(img_name: str) -> pathlib.Path:
    """Capture image with `label` appended to filename, and return the filepath."""
    if RIG.idx == 0:
        from np_services import Cam3d as ImageCamera
    else:
        from np_services import ImageMVR as ImageCamera
    from np_services import NewScaleCoordinateRecorder
    
    ImageCamera.label = img_name
    if isinstance(ImageCamera, Initializable) and not getattr(ImageCamera, 'initialization', None):
        ImageCamera.initialize()
        
    ImageCamera.start()
    
    if isinstance(ImageCamera, Finalizable):
        ImageCamera.finalize()
        
    NewScaleCoordinateRecorder.label = img_name
    NewScaleCoordinateRecorder.start()
    return ImageCamera.data_files[-1]

def copy_files(services: Sequence[Service], session_folder: pathlib.Path):
    """Copy files from raw data storage to session folder for all services."""
    password = input(f'Enter password for svc_neuropix:')
    for service in services:
        match service.__class__.__name__:
            case "OpenEphys" | "open_ephys":
                continue # copy ephys after other files
            case _:
                with contextlib.suppress(AttributeError):
                    files = service.data_files or service.get_latest_data('*')
                    if not files:
                        continue
                    files = set(files)
                    print(files)
                    for file in files:
                        shutil.copy2(file, session_folder)
                    
    # copy ephys                    
    for ephys_folder in OpenEphys.data_files:
        fabric.Connection(host=OpenEphys.host, user='svc_neuropix', connect_kwargs=dict(password=password)).run(
            f'robocopy "{ephys_folder}" "{session_folder}" /j /s /xo' # /j unbuffered, /s incl non-empty subdirs, /xo exclude src files older than dest
        ) 
            
            
import warnings

def hide_warning_lines(msg:str,category:str,*args,**kwargs):
    print("\n{}: {}\n".format(category.__name__, msg))
warnings.showwarning = hide_warning_lines

import IPython
def toggle_tracebacks() -> Generator[None, None, None]:
    if ipython := IPython.get_ipython():
        show_traceback = ipython.showtraceback
        
        def hide_traceback(exc_tuple=None, filename=None, tb_offset=None,
                        exception_only=False, running_compiled_code=False):
            etype, value, tb = sys.exc_info()
            return ipython._showtraceback(etype, value, ipython.InteractiveTB.get_exception_only(etype, value))
        
        hidden = True
        while True:
            ipython.showtraceback = hide_traceback if hidden else show_traceback
            hidden = yield
    else:
        raise RuntimeError("Not in IPython")
    
with contextlib.suppress(RuntimeError):
    toggle_tb = toggle_tracebacks()

    toggle_tb.send(None)
    def show_tracebacks():
        toggle_tb.send(False)
    def hide_tracebacks():
        toggle_tb.send(True)
        
def now() -> str:
    return np_services.utils.normalize_time(time.time())


from __future__ import annotations

import datetime
import pathlib
import typing
from typing import Any, ClassVar, Optional, Protocol, Sequence, Type, Union

import np_config
import np_session

from np_services.protocols import Service

from np_workflows.models.experiment import WithLims

RESUME_ROOT = pathlib.Path(
    np_config.from_zk("/projects/np_workflows/defaults/configuration")[
        "workflow_resume_root"
    ]
)
RESUME_DATETIME_FMT = "%Y-%m-%d_%H%M%S"


def save(experiment: WithLims, suffix: Optional[str] = None):
    path = RESUME_ROOT / experiment.__class__.__name__ / experiment.session
    filename = datetime.datetime.now().strftime(RESUME_DATETIME_FMT)
    if suffix:
        filename = f"{filename}_{suffix}"
    path.mkdir(exist_ok=True, parents=True)
    pickle.dump(experiment, path / f"{filename}.pkl")


def sorted_dirs(path: pathlib.Path) -> list[pathlib.Path]:
    return sorted(
        [p for p in path.iterdir() if p.is_dir()],
        key=lambda _: _.stat().ctime,
        reverse=True,
    )


def most_recent_session(experiment_type: Type[WithLims]) -> pathlib.Path | None:
    root = RESUME_ROOT / experiment_type.__name__
    if root.exists() and (dirs := sorted_dirs(root)):
        return dirs[0]
    return None


def saves_from_most_recent_session(
    experiment_type: Type[WithLims],
) -> list[pathlib.Path] | None:
    if session_dir := most_recent_session(experiment_type):
        return sorted_dirs(session_dir)
    return None


def most_recent_save(experiment_type: Type[WithLims]) -> pathlib.Path | None:
    "Path to most recent save state"
    if previous := saves_from_most_recent_session(experiment_type):
        return sorted(previous.glob("*.pkl"), key=lambda _: _.name)
    return None


def load(path: pathlib.Path):
    return pickle.load(path)



# -*- coding: utf-8 -*-
"""
optotagging.py

runs optotagging code for ecephys pipeline experiments

by joshs@alleninstitute.org

(c) 2018 Allen Institute for Brain Science

"""
import camstim  # ensures "magic" gets setup properly by importing first
import logging  # must occur after camstim import for "magic"
from camstim.zro import agent

import numpy as np
from toolbox.IO.nidaq import AnalogOutput
from toolbox.IO.nidaq import DigitalOutput

import datetime
import numpy as np
import time
import pickle as pkl


# %%


def run_optotagging(levels, conditions, waveforms, isis, sampleRate=10000.):

    from toolbox.IO.nidaq import AnalogOutput
    from toolbox.IO.nidaq import DigitalOutput

    sweep_on = np.array([0, 0, 1, 0, 0, 0, 0, 0], dtype=np.uint8)
    stim_on = np.array([0, 0, 1, 1, 0, 0, 0, 0], dtype=np.uint8)
    stim_off = np.array([0, 0, 1, 0, 0, 0, 0, 0], dtype=np.uint8)
    sweep_off = np.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=np.uint8)

    ao = AnalogOutput('Dev1', channels=[1])
    ao.cfg_sample_clock(sampleRate)

    do = DigitalOutput('Dev1', 2)

    do.start()
    ao.start()

    do.write(sweep_on)
    time.sleep(5)

    for i, level in enumerate(levels):

        print(level)

        data = waveforms[conditions[i]]

        do.write(stim_on)
        ao.write(data * level)
        do.write(stim_off)
        time.sleep(isis[i])

    do.write(sweep_off)
    do.clear()
    ao.clear()

# %%


def generatePulseTrain(pulseWidth, pulseInterval, numRepeats, riseTime, sampleRate=10000.):

    data = np.zeros((int(sampleRate),), dtype=np.float64)
   # rise_samples =

    rise_and_fall = (
        ((1 - np.cos(np.arange(sampleRate*riseTime/1000., dtype=np.float64)*2*np.pi/10))+1)-1)/2
    half_length = rise_and_fall.size / 2
    rise = rise_and_fall[:half_length]
    fall = rise_and_fall[half_length:]

    peak_samples = int(sampleRate*(pulseWidth-riseTime*2)/1000)
    peak = np.ones((peak_samples,))

    pulse = np.concatenate((rise,
                            peak,
                            fall))

    interval = int(pulseInterval*sampleRate/1000.)

    for i in range(0, numRepeats):
        data[i*interval:i*interval+pulse.size] = pulse

    return data


# %% create waveforms

def optotagging(mouseID, operation_mode='experiment', level_list=[1.15, 1.28, 1.345], genotype=None):

    sampleRate = 10000

    # 1 s cosine ramp:
    data_cosine = (((1 - np.cos(np.arange(sampleRate, dtype=np.float64)
                                * 2*np.pi/sampleRate)) + 1) - 1)/2  # create raised cosine waveform

    # 1 ms cosine ramp:
    rise_and_fall = (
        ((1 - np.cos(np.arange(sampleRate*0.001, dtype=np.float64)*2*np.pi/10))+1)-1)/2
    half_length = rise_and_fall.size / 2

    # pulses with cosine ramp:
    pulse_2ms = np.concatenate((rise_and_fall[:half_length], np.ones(
        (int(sampleRate*0.001),)), rise_and_fall[half_length:]))
    pulse_5ms = np.concatenate((rise_and_fall[:half_length], np.ones(
        (int(sampleRate*0.004),)), rise_and_fall[half_length:]))
    pulse_10ms = np.concatenate((rise_and_fall[:half_length], np.ones(
        (int(sampleRate*0.009),)), rise_and_fall[half_length:]))

    data_2ms_10Hz = np.zeros((sampleRate,), dtype=np.float64)

    for i in range(0, 10):
        interval = sampleRate / 10
        data_2ms_10Hz[i*interval:i*interval+pulse_2ms.size] = pulse_2ms

    data_5ms = np.zeros((sampleRate,), dtype=np.float64)
    data_5ms[:pulse_5ms.size] = pulse_5ms

    data_10ms = np.zeros((sampleRate,), dtype=np.float64)
    data_10ms[:pulse_10ms.size] = pulse_10ms

    data_10s = np.zeros((sampleRate*10,), dtype=np.float64)
    data_10s[:-2] = 1

    # %% for experiment

    isi = 1.5
    isi_rand = 0.5
    numRepeats = 50

    condition_list = [2, 3]
    waveforms = [data_2ms_10Hz, data_5ms, data_10ms, data_cosine]

    opto_levels = np.array(level_list*numRepeats*len(condition_list))  # BLUE
    opto_conditions = condition_list*numRepeats*len(level_list)
    opto_conditions = np.sort(opto_conditions)
    opto_isis = np.random.random(opto_levels.shape) * isi_rand + isi

    p = np.random.permutation(len(opto_levels))

    # implement shuffle?
    opto_levels = opto_levels[p]
    opto_conditions = opto_conditions[p]

    # %% for testing

    if operation_mode == 'test_levels':
        isi = 2.0
        isi_rand = 0.0

        numRepeats = 2

        condition_list = [0]
        waveforms = [data_10s, data_10s]

        opto_levels = np.array(level_list*numRepeats *
                               len(condition_list))  # BLUE
        opto_conditions = condition_list*numRepeats*len(level_list)
        opto_conditions = np.sort(opto_conditions)
        opto_isis = np.random.random(opto_levels.shape) * isi_rand + isi

    elif operation_mode == 'pretest':
        numRepeats = 1

        condition_list = [0]
        data_2s = data_10s[-sampleRate*2:]
        waveforms = [data_2s]

        opto_levels = np.array(level_list*numRepeats *
                               len(condition_list))  # BLUE
        opto_conditions = condition_list*numRepeats*len(level_list)
        opto_conditions = np.sort(opto_conditions)
        opto_isis = [1]*len(opto_conditions)
    # %%

    outputDirectory = agent.OUTPUT_DIR
    fileDate = str(datetime.datetime.now()).replace(':', '').replace(
        '.', '').replace('-', '').replace(' ', '')[2:14]
    fileName = outputDirectory  + "/" + fileDate + '_'+mouseID + '.opto.pkl'

    print('saving info to: ' + fileName)
    fl = open(fileName, 'wb')
    output = {}

    output['opto_levels'] = opto_levels
    output['opto_conditions'] = opto_conditions
    output['opto_ISIs'] = opto_isis
    output['opto_waveforms'] = waveforms

    pkl.dump(output, fl)
    fl.close()
    print('saved.')

    # %%
    run_optotagging(opto_levels, opto_conditions,
                    waveforms, opto_isis, float(sampleRate))


# %%
if __name__ == "__main__":
    import json
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('json_params', type=str, )
    args, _ = parser.parse_known_args()

    with open(args.json_params, 'r', ) as f:
        json_params = json.load(f)

    logging.info('Optotagging with params: %s' % json_params)
    optotagging(**json_params)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['np_workflows',
 'np_workflows.experiments',
 'np_workflows.experiments.task_trained_network',
 'np_workflows.experiments.task_trained_network.camstim_scripts',
 'np_workflows.models',
 'np_workflows.shared']

package_data = \
{'': ['*'], 'np_workflows': ['assets/images/*']}

install_requires = \
['ipylab>=0.6.0,<0.7.0',
 'ipywidgets>=7,<8',
 'jupyter-scheduler>=1.2.0,<2.0.0',
 'jupyterlab-git>=0.41.0,<0.42.0',
 'jupyterlab>=3.6,<4.0',
 'np_datajoint',
 'np_probe_targets',
 'np_services>=0.1.33',
 'pydantic>=1,<2']

setup_kwargs = {
    'name': 'np-workflows',
    'version': '1.3.18',
    'description': 'Ecephys and behavior workflows for the Mindscope Neuropixels team.',
    'long_description': '# np_workflows',
    'author': 'Ben Hardcastle',
    'author_email': 'ben.hardcastle@alleninstitute.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)



"""
Oct'22 task-trained ephys stimuli
"""

import argparse
import json
import logging
import os
import time

import numpy as np
from psychopy import visual
from camstim import Foraging
from camstim import Stimulus_v2
from camstim import SweepStim_v2
from camstim import Warp, Window


# get params ------------------------------------------------------------------
# stored in json file -
# path to json supplied by camstim via command line arg when this script is called

parser = argparse.ArgumentParser()
parser.add_argument(
    "params_path",
    nargs="?",
    type=str,
    default="",
)
args, _ = parser.parse_known_args()

with open(args.params_path, "r") as f:
    json_params = json.load(f)

# Create display window
# ----------------------------------------------------------------------------
window = Window(
    fullscr=True,
    monitor=json_params["monitor"],
    screen=0,
    warp=Warp.Disabled,
)

# patch the Stimulus_v2 class to allow for serializing without large arrays
# ----------------------------------------------------------------------------
class Stimulus_v2_MinusFrameArrays(Stimulus_v2):

    def __init__(self, *args, **kwargs):
        super(Stimulus_v2_MinusFrameArrays, self).__init__(*args, **kwargs)

    def package(self):
        """
        Package for serializing - minus large arrays of frame timing/order.
        """
        if not self.save_sweep_table:
            self.sweep_table = None
            self.sweep_params = self.sweep_params.keys()
        self_dict = self.__dict__
        del self_dict['sweep_frames']
        del self_dict['sweep_order']
        self_dict['stim'] = str(self_dict['stim'])
        return wecanpicklethat(self_dict)

# setup main stim
# -----------------------------------------------------------------------
# build the stimulus array with parameterized repeats & durations

main_stimuli = []

old_repeats, reversed_repeats, annotated_repeats = (
    json_params["stim_repeats"][key] for key in ("old", "reversed", "annotated")
)
old_sec, reversed_sec, annotated_sec = (
    json_params["stim_lengths_sec"][key] for key in ("old", "reversed", "annotated")
)

segment_stim_secs = (
    [("old_stim.stim", old_sec)] * old_repeats
    + [
        ("shuffle_reversed.stim", reversed_sec),
        ("shuffle_reversed_1st.stim", reversed_sec),
        ("shuffle_reversed_2nd.stim", reversed_sec),
    ] * reversed_repeats
    + [("densely_annotated_%02d.stim" % i, annotated_sec) for i in range(19)]
    * annotated_repeats
    + [("old_stim.stim", old_sec)] * old_repeats
    + [
        ("shuffle_reversed.stim", reversed_sec),
        ("shuffle_reversed_1st.stim", reversed_sec),
        ("shuffle_reversed_2nd.stim", reversed_sec),
    ] * reversed_repeats
)

# setup stim list and timing
cumulative_duration_sec = (
    main_sequence_start_sec
) = 0  # if stims are daisy-chained within one script, this should be the end of the prev stim
for stim_file, duration_sec in segment_stim_secs:
    segment = Stimulus_v2_MinusFrameArrays.from_file(stim_file, window)
    segment_ds = [(cumulative_duration_sec, cumulative_duration_sec + duration_sec)]
    segment.set_display_sequence(segment_ds)

    cumulative_duration_sec += duration_sec
    main_stimuli.append(segment)

main_sequence_end_sec = cumulative_duration_sec  # if daisy-chained, the next stim in this script should start at this time

# create SweepStim_v2 instance for main stimulus
ss = SweepStim_v2(
    window,
    stimuli=main_stimuli,
    pre_blank_sec=json_params["pre_blank_screen_sec"],
    post_blank_sec=json_params["post_blank_screen_sec"],
    params=json_params["sweepstim"],
)

# add in foraging so we can track wheel, potentially give rewards, etc
f = Foraging(
    window=window,
    auto_update=False,
    params=json_params["sweepstim"],
    nidaq_tasks={
        "digital_input": ss.di,
        "digital_output": ss.do,
    },
)  # share di and do with SS

ss.add_item(f, "foraging")

ss.run()



import abc
import contextlib
import functools
import getpass
import pathlib
import re
import shutil
import time
from typing import Any, ClassVar, Iterable, Literal, Optional, Protocol, Sequence, Type, Union

import fabric 
import ipylab

import np_config
import np_logging
import np_session
import np_services
import np_workflows.npxc as npxc
from np_services import (
    Service,
    TestError,
    Configurable,
    Initializable,
    Testable,
    Pretestable,
    Startable,
    Verifiable,
    Stoppable,
    Finalizable,
    Validatable,
    Shutdownable,
)


logger = np_logging.getLogger(__name__)


class WithLims(abc.ABC):
    
    default_session_type: Literal['ecephys', 'hab'] = 'ecephys'
    
    services: tuple[Service, ...] = ()
    "Devices, databases, etc."

    def __init__(self, 
        mouse: Optional[str | int |  np_session.LIMS2MouseInfo] = None,
        operator: Optional[str | np_session.LIMS2UserInfo] = None, 
        session: Optional[str | pathlib.Path | int | np_session.LIMS2SessionInfo] = None,
        session_type: Literal['ecephys', 'hab'] = default_session_type,
        **kwargs,
        ):
        
        # np_config.merge(self.__dict__, kwargs)
        
        if session:
            self.session = session
        elif operator and mouse:
            self.session = np_session.generate_session(mouse, operator, session_type)
        else:
            raise ValueError('Must specify either a mouse + operator, or an existing session')
    
        self.session_type = session_type
            
        self.configure_services()
        self.session.npexp_path.mkdir(parents=True, exist_ok=True)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.session})'
    
    @property
    def platform_json(self) -> np_services.PlatformJsonWriter:
        with contextlib.suppress(AttributeError):
            return self._platform_json
        self._platform_json = np_services.PlatformJsonWriter(path=self.session.npexp_path)
        self._platform_json.operatorID = str(self.user)
        self._platform_json.mouseID = str(self.mouse)
        self._platform_json.sessionID = str(self.session)
        return self.platform_json
    
    @property
    def session(self) -> np_session.Session:
        return self._session
        
    @session.setter
    def session(self, value: str | np_session.Session | pathlib.Path | int | np_session.LIMS2SessionInfo):
        self._session = np_session.Session(value) if not isinstance(value, np_session.Session) else value
        logger.debug('Set experiment.session to %r', self._session)
    
    @property
    def session_type(self) -> Literal['ecephys', 'hab']:
        with contextlib.suppress(AttributeError):
            return self._session_type
        if self.session:
            if self.session.is_ecephys_session:
                return 'ecephys'
            return 'hab'
        raise AttributeError('Session has not been set')
    
    @session_type.setter
    def session_type(self, value: Literal['ecephys', 'hab']):
        if value not in ('ecephys', 'hab'):
            raise ValueError('Session type must be either "ecephys" or "hab"')
        self._session_type = value
        logger.debug('Set session_type to %r', value)
        
    @property
    def rig(self) -> np_config.Rig | None:
        "Computer hostnames and configuration for the rig we're currently on."
        with contextlib.suppress(AttributeError):
            return self._rig
        with contextlib.suppress(ValueError):
            self._rig = np_config.Rig()
            return self.rig
    
    @property
    def mouse(self) -> np_session.Mouse:
        return self.session.mouse
    
    @property
    def user(self) -> np_session.User | None:
        return self.session.user
    
    @property
    def imager(self) -> Type[np_services.Cam3d] | Type[np_services.ImageMVR]:
        if not self.rig:
            raise ValueError('Rig has not been set')
        if self.rig.idx == 0:
            return np_services.Cam3d
        return np_services.ImageMVR
    
    @property
    def config(self) -> dict[Any, Any]:
        "Top-level keys include names of Services. Each Service then has a config dict."
        with contextlib.suppress(AttributeError):
            return self._config
        if self.rig:
            self._config = self.rig.config
        return self.config
        
    def configure_services(self) -> None:
        """For each service, apply every key in self.config['service'] as an attribute."""

        def apply_config(service):
            if config := self.config["services"].get(service.__name__):
                for key, value in config.items():
                    setattr(service, key, value)
                    logger.debug(
                        f"{self.__class__.__name__} | Configuring {service.__name__}.{key} = {getattr(service, key)}"
                    )

        for service in self.services:
            for base in service.__class__.__bases__:
                apply_config(base)
            apply_config(service)

    def initialize_and_test_services(self) -> None:

        for service in self.services:

            if isinstance(service, Initializable):
                service.initialize()

            if isinstance(service, Testable):
                service.test()

    def pretest_services(self) -> None:
        for service in (_ for _ in self.services if isinstance(_, Pretestable)):
            service.pretest()
    
    def start_recording(self, *recorders: Startable) -> None:
        if not recorders and hasattr(self, 'recorders'):
            recorders = self.recorders
        stoppables = tuple(_ for _ in recorders if isinstance(_, Stoppable))
        with np_services.stop_on_error(*stoppables):
            for recorder in recorders:
                recorder.start()
                time.sleep(2)
                if isinstance(recorder, Verifiable):
                    recorder.verify()
        self.platform_json.ExperimentStartTime = npxc.now()        
        self.platform_json.write()
    
    def stop_recording_after_stim_finished(self, 
            recorders: Optional[Iterable[Startable]] = None, 
            stims: Optional[Iterable[Startable]] = None,
        ) -> None:
        """Stop recording after all stims have finished."""
        if not recorders and hasattr(self, 'recorders'):
            recorders = self.recorders
        if not stims and hasattr(self, 'stims'):
            stims = self.stims
        while not all(_.is_ready_to_start() for _ in stims):
            time.sleep(5)
        for stoppable in (_ for _ in recorders if isinstance(_, Stoppable)):
            stoppable.stop()
        self.platform_json.ExperimentCompleteTime = npxc.now()        
        self.platform_json.write()
                
    def start_services(self, *services: Service) -> None:
        if not services:
            services = self.services
        for service in (_ for _ in services if isinstance(_, Startable)):
            service.start()
            if isinstance(service, Verifiable):
                service.verify()

    def stop_services(self) -> None:
        while not np_services.ScriptCamstim.is_ready_to_start():
            time.sleep(10)
        for service in (_ for _ in self.services if isinstance(_, Stoppable)):
            service.stop()
            if isinstance(service, Finalizable):
                service.finalize()

    def validate_services(self, *services: Service) -> None:
        if not services:
            services = self.services
        for service in (_ for _ in services if isinstance(_, Validatable)):
            service.validate()
            
    def finalize_services(self, *services: Service) -> None:
        if not services:
            services = self.services
        for service in (_ for _ in services if isinstance(_, Finalizable)):
            service.finalize()

    def shutdown_services(self) -> None:
        for service in (_ for _ in self.services if isinstance(_, Shutdownable)):
            service.shutdown()
            
    def copy_files(self):
        """Copy files from raw data storage to session folder for all services."""
        session_folder = self.session.npexp_path
        for service in self.services:
            match service.__name__:
                case "np_services.open_ephys":
                    continue # copy ephys after other files
                case _:
                    with contextlib.suppress(AttributeError):
                        files = service.data_files or service.get_latest_data('*')
                        if not files:
                            continue
                        files = set(files)
                        logger.info("%s | Copying files %r", service.__name__, files)
                        for file in files:
                            if file.suffix == '.h5':
                                renamed = f'{self.session.folder}.sync'
                            elif file.suffix == '.pkl':
                                for _ in ('opto', 'main', 'mapping'):
                                    if _ in file.name:
                                        renamed = f'{self.session.folder}.{"stim" if _ == "main" else _}.pkl'
                            elif file.suffix in ('.json', '.mp4') and (cam_label := re.match('Behavior|Eye|Face',file.name)):
                                renamed = f'{self.session.folder}.{cam_label.group().lower()}{file.suffix}'
                            shutil.copy2(file, session_folder / renamed)

        # copy ipynb, logs and lock/pyproject files
        app = ipylab.JupyterFrontEnd()
        app.commands.execute('docmanager:save')

        src = pathlib.Path('.').resolve()
        dest = self.session.npexp_path / 'exp'
        dest.mkdir(exist_ok=True, parents=True)

        ipynb = sorted(src.glob('*.ipynb'), key=lambda path: path.stat().st_mtime, reverse=True)[0]
        logs = src / 'logs'
        lock = src.parent / 'pdm.lock'
        pyproject = src.parent / 'pyproject.toml'

        shutil.copytree(src, dest, dirs_exist_ok=True)
        for _ in (lock, pyproject):
            shutil.copy2(_, dest)
            
        # copy ephys       
        password = getpass.getpass(f'Enter password for svc_neuropix:')
        ssh = fabric.Connection(host=np_services.OpenEphys.host, user='svc_neuropix', connect_kwargs=dict(password=password))
        for ephys_folder in np_services.OpenEphys.data_files:

            if ephys_folder.drive.endswith("A"):
                probes = '_probeABC'
            elif ephys_folder.drive.endswith("B"):
                probes = '_probeDEF'
            else:
                probes = '_probes'
                
            with contextlib.suppress(Exception):
                with ssh:
                    ssh.run(
                    f'robocopy "{ephys_folder}" "{session_folder / (session_folder.name + probes)}" /j /s /xo' 
                    # /j unbuffered, /s incl non-empty subdirs, /xo exclude src files older than dest
                    ) 
                           
    def photodoc(self, img_label: Optional[str] = None) -> pathlib.Path:
        """Capture image with `img_label` appended to filename, and return the filepath."""        
        if img_label:
            self.imager.label = img_label
        
        if isinstance(self.imager, Initializable) and not getattr(self.imager, 'initialization', None):
            self.imager.initialize()
            
        self.imager.start()
        
        if isinstance(self.imager, Finalizable):
            self.imager.finalize()
            
        if (recorder := np_services.NewScaleCoordinateRecorder) in self.services:
            if img_label:
                recorder.label = img_label
            recorder.start()
            
        return self.imager.data_files[-1]
    
class Ecephys(WithLims):
    default_session_type = 'ecephys'

class Hab(WithLims):
    default_session_type = 'hab'
